### Produces Table 3 ("Moments of discretized Lévy processes for various period durations")

### Also produces PDFs for value, policy and price functions

library(stats)
library(statmod)
library(parallel)

### Brownian and Poisson observationally equivalent at h=h0

sigma <- 0.022
mu <- 0.01
h0 <- 4

### 1/(4*sigma+mu/sigma)^2 = 3.40

num.cores <- detectCores()

simulate.moments <- function(h, brownian, min.a = -0.8, nsims = 5000)
{
  gam <- 2
  lambda <- 0.005
  beta <- 0.8^h
  r <- 1.01^h-1
  
  if (brownian) {
    gu <- exp(sigma*sqrt(h))
    gd <- 1/gu
    pu <- 0.5 + mu/2/sigma*sqrt(h)
    pd <- 1-pu
  } else {
    p0 <- -log(0.5 + mu/2/sigma*sqrt(h0))/h0
    
    gd <- exp(-sigma*sqrt(h0))/(p0*h0/(1-exp(-p0*h0)))*(p0*h/(1-exp(-p0*h)))
    pu <- exp(-p0*h)
    pd <- 1-exp(-p0*h)

    ## Compute gu so that pu*gu+pd*gd the same than in the Brownian of same h
    x.brownian <- (0.5 + mu/2/sigma*sqrt(h))*exp(sigma*sqrt(h)) + (0.5 - mu/2/sigma*sqrt(h))*exp(-sigma*sqrt(h))
    gu <- (x.brownian - pd*gd)/pu
  }

  stopifnot(pu > 0 && pd > 0)

  stopifnot(beta*(pu*gu^(1-gam) + pd*gd^(1-gam))*(1+r) < 1)

  max.a <- 0
  if (h == 0.08)
    ngrid.a <- 40
  else
    ngrid.a <- 200
  convergence.tol <- 1e-5

  VG <- NULL
  VB <- NULL
  threshold <- NA

  ## Grids
  grid.a <- seq(min.a, max.a, length.out = ngrid.a)

  util <- function(c)
    ifelse(c <= 0, -1e8, c^(1-gam)/(1-gam))

  V <- function(a)
    max(VB, VG(a))

  q <- function(a1)
    {
      if (a1 >= 0)
        return(1/(1+r))
      
      Erepay1 <- 0

      if (VG(a1/gu) >= VB)
        Erepay1 <- Erepay1 + pu

      if (VG(a1/gd) >= VB)
        Erepay1 <- Erepay1 + pd
      
      Erepay1/(1+r)
    }

  a1.dr <- function(a)
    {
      VG.objective <- function(a1)
        util(h+a-a1*q(a1)) + beta*(pu*gu^(1-gam)*V(a1/gu) + pd*gd^(1-gam)*V(a1/gd))
      
      zd <- max(min(threshold*gd, max.a), min.a)
      o1 <- optimize(VG.objective, lower = min.a, upper = zd, maximum = TRUE, tol = convergence.tol)
      o2 <- optimize(VG.objective, lower = zd, upper = max.a, maximum = TRUE, tol = convergence.tol)

      ifelse(o1$objective > o2$objective, o1$maximum, o2$maximum)
    }

  solve <- function(max.it = 10000)
    {
      ## Initial guess
      VB <<- util((1-lambda)*h)
      VG.points <- util(h + grid.a*r)
      threshold <<- min(max(-lambda*h/r, min.a), max.a)

      init.spline <- function()
        VG <<- splinefun(grid.a, VG.points, method = "monoH.FC")

      init.spline()

      ## Run iterations
      
      it <- 1
      err <- Inf

      VG.points.new <- numeric(ngrid.a)

      while(err > convergence.tol && it <= max.it)
        {
          difference <- function(a)
            VG(a)-VB

          VB.new <- util((1-lambda)*h) + beta*(pu*gu^(1-gam) + pd*gd^(1-gam))*VB
          
          ## Compute VG
          VG.optimize <- function(a)
            {
              VG.objective <- function(a1)
                util(h+a-a1*q(a1)) + beta*(pu*gu^(1-gam)*V(a1/gu) + pd*gd^(1-gam)*V(a1/gd))

              zd <- max(min(threshold*gd, max.a), min.a)
              o1 <- optimize(VG.objective, lower = min.a, upper = zd, maximum = TRUE, tol = convergence.tol)
              o2 <- optimize(VG.objective, lower = zd, upper = max.a, maximum = TRUE, tol = convergence.tol)

              max(o1$objective, o2$objective)
            }
          VG.points.new <- as.numeric(mclapply(grid.a, VG.optimize, mc.cores = num.cores))

          ## Compute error
          err <- max(abs(VB.new-VB),
                     max(abs(VG.points.new-VG.points)))

          ## Copy
          VB <<- VB.new
          VG.points <- VG.points.new
          init.spline()

          ## Threshold
          if (difference(min.a) < 0 && difference(max.a) > 0)
            {
              threshold <<- uniroot(difference, lower = min.a, upper = max.a, tol = convergence.tol)$root
              cat(threshold, fill = TRUE)
            }
          else if (difference(min.a) >= 0)
            threshold <<- min.a
          else
            threshold <<- max.a


          ## Output
          cat("Iteration ", it, ": error=", err, fill = TRUE, sep = "")

          it <- it+1
        }

      if (threshold > a1.dr(threshold)/gd + convergence.tol)
        cat("Models with defaults", fill = TRUE)
      else
        cat("No defaults in this model", fill = TRUE)
      
      if (it == max.it+1)
        warning("Maximum iterations reached")

      threshold
    }

  simulation.has.default <- function(a0, length)
    {
      a <- a0
      
      for (i in 1:length)
        {
          if (VG(a) >= VB)
            {
              a <- a1.dr(a)/ifelse(runif(1) < pu, gu, gd)
            }
          else
            {
              return(TRUE)
            }
        }
      FALSE
    }

  default.prob <- function(nsims, sim.length = 40/h)
    {
      count <- 0
      for (i in 1:nsims)
        if (simulation.has.default(0, sim.length))
          count <- count + 1
      count / nsims
    }
  
  t <- solve()

  if (nsims > 0)
    d <- mean(pvec(rep(NA, num.cores), function(x) default.prob(nsims = ceiling(nsims/num.cores)), mc.cores = num.cores))
  else
    d <- NA
  
  list(threshold = t, default.prob = d, VG = VG, VB = VB, a1.dr = a1.dr, q = q, gu = gu, gd = gd)
}

find.discontinuities <- function(f, min.x, max.x, nb.points, deriv.threshold = 100)
{
  r <- list()
  grid <- seq(from = min.x, to = max.x, length.out = nb.points)
  for (i in 1:(length(grid)-1))
  {
    a <- grid[i]
    b <- grid[i+1]
    if (abs((f(a)-f(b))/(b-a)) >= deriv.threshold)
      r <- c(r, list(c(a,b)))
  }
  r
}

h <- c(h0, 2, 1, 0.33, 0.08)
min.a <- c(-0.8, -0.8, -0.7, -0.8, -0.6, -0.8, -0.6, -0.9, -0.6, -0.5)

results <- list()

for (i in 1:length(min.a))
  {
    this.h <- h[(i-1) %/% 2 + 1]
    this.brownian <- as.logical((i-1) %% 2)
    l <- simulate.moments(this.h, this.brownian, min.a[i])
    l[["h"]] <- this.h
    l[["brownian"]] <- this.brownian
    results[[i]] <- l
  }

sink("vfi-2states.tex")
cat("\\begin{tabular}{l", paste(rep("|r", length(min.a)), collapse = ""), "}", fill = TRUE)
cat("\\hline \\hline", fill = TRUE)
cat("Period duration ($h$, in quarters) & ", paste(round(h, digit = 2), collapse = " & "), "\\\\ \\hline", fill = TRUE)
threshold.select <- function(i) formatC(abs(results[[i]]$threshold*100), format = "f", digits = 1)
default.prob.select <- function(i) formatC(results[[i]]$default.prob*100, format = "f", digits = 1)
cat("\\textit{Discretized Brownian process}", paste(rep("&", length(h))), "\\\\", fill = TRUE)
cat("Default threshold (debt-to-GDP, quarterly, \\%) & ", paste(sapply((1:length(h))*2, threshold.select), collapse = " & "), "\\\\", fill = TRUE)
cat("Default probability in 10 years (\\%) & ", paste(sapply((1:length(h))*2, default.prob.select), collapse = " & "), "\\\\ \\hline", fill = TRUE)
cat("\\textit{Discretized Poisson process}", paste(rep("&", length(h))), "\\\\", fill = TRUE)
cat("Default threshold (debt-to-GDP, quarterly, \\%) & ", paste(sapply((1:length(h))*2-1, threshold.select), collapse = " & "), "\\\\", fill = TRUE)
cat("Default probability in 10 years (\\%) & ", paste(sapply((1:length(h))*2-1, default.prob.select), collapse = " & "), "\\\\ \\hline \\hline", fill = TRUE)
cat("\\end{tabular}", fill = TRUE)
sink()


## Graphs of value, policy and price functions

l <- simulate.moments(1, TRUE, -0.8, nsims = 0)
pdf(file = "Value-fct-Brownian.pdf", width = 4, height = 4, pointsize = 8)
plot(function(x) l$VG(-x), from = 0, to = 0.8, xlab = "D/Y", ylab = "J", main = "Value functions for discretized Brownian (h=1)")
abline(h=l$VB, lty = 2)
legend(x = "topright", lty = c(1,2), legend = c("Jr", "Jd"))
dev.off()
pdf(file = "Policy-fct-Brownian.pdf", width = 4, height = 4, pointsize = 8)
plot(function(x) -sapply(-x, l$a1.dr), from = 0, to = 0.8, n = 500, xlab = "D/Y", ylab = "D'/Y", main = "Policy function for discretized Brownian (h=1)")
abline(v=-l$t, lty = 2)
abline(a=0, b=l$gu, lty = 3)
abline(a=0, b=l$gd, lty = 4)
dev.off()
pdf(file = "Price-fct-Brownian.pdf", width = 4, height = 4, pointsize = 8)
plot(function(x) sapply(-x, l$q), from = 0, to = 0.8, n = 500, xlab = "D'/Y", ylab = "q", main = "Price schedule for discretized Brownian (h=1)")
abline(v=-l$t, lty = 2)
dev.off()

l <- simulate.moments(1, FALSE, -0.6, nsims = 0)
pdf(file = "Value-fct-Poisson.pdf", width = 4, height = 4, pointsize = 8)
plot(function(x) l$VG(-x), from = 0, to = 0.6, xlab = "D/Y", ylab = "J", main = "Value functions for discretized Poisson (h=1)")
abline(h=l$VB, lty = 2)
legend(x = "topright", lty = c(1,2), legend = c("Jr", "Jd"))
dev.off()
pdf(file = "Policy-fct-Poisson.pdf", width = 4, height = 4, pointsize = 8)
pol.fct <- function(x) -sapply(-x, l$a1.dr)
disc <- find.discontinuities(pol.fct, 0, 0.6, 10000)
## Remove 1st discontinuity, it may just be a very fast moving objective
disc[[1]] <- NULL
for (i in 1:length(disc))
  cat("Discontinuity in x=", mean(disc[[i]]), " from y=", pol.fct(disc[[i]][1]), " to y=", pol.fct(disc[[i]][2]), fill = TRUE)
disc <- c(0, unlist(disc), 0.6)
plot(pol.fct, from = 0, to = 0.6, n = 500, xlab = "D/Y", ylab = "D'/Y", main = "Policy function for discretized Poisson (h=1)", type = "n")
for (i in seq(from=1, to=length(disc), by = 2))
  plot(pol.fct, from = disc[i], to = disc[i+1], n = 500, add = TRUE)
abline(v=-l$t, lty = 2)
abline(a=0, b=l$gu, lty = 3)
abline(a=0, b=l$gd, lty = 4)
dev.off()
pdf(file = "Price-fct-Poisson.pdf", width = 4, height = 4, pointsize = 8)
plot(function(x) sapply(-x, l$q), from = 0, to = 0.6, n = 500, xlab = "D'/Y", ylab = "q", main = "Price schedule for discretized Poisson (h=1)")
abline(v=-l$t, lty = 2)
dev.off()
